# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""
nose tests
"""

from ..dev._helpers import unicode_to_utf8tex
from nose.tools import assert_equals


def test_unicode_to_utf8tex():
    """Test 'unicode_to_utf8tex' identity transformation."""
    assert_equals(unicode_to_utf8tex(u"\u00E2 \u00F5"), b"\^{a}\space \~{o}")
    assert_equals(unicode_to_utf8tex(u"\u00E2 \u00F5", (b"\\space ",)), b"\^{a} \~{o}")



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""
nose tests
"""

import os

from .. import _merge as merge
from nose.tools import assert_equals


WINDOWS = os.name == 'nt'


def test_vias_modes():
    """Test 'vias' several modes of operation."""
    if WINDOWS:
        # appveyor doesn't allow!
        return
    assert_equals(len(repr(merge.query('9780321534965', 'parallel'))) > 170, True)
    assert_equals(len(repr(merge.query('9780321534965', 'multi'))) > 170, True)
    assert_equals(len(repr(merge.query('9780321534965', 'serial'))) > 170, True)

def test_vias_cache_cleanning():
    """Test 'vias' cache cleanning for serial."""
    # test if the secondary cache (cache in vias) does clears... sequentially
    assert_equals(len(repr(merge.query('9781484206546', 'serial'))) < 20, True)  # NO METADATA
    assert_equals(len(repr(merge.query('9780321534965', 'serial'))) > 170, True)
    assert_equals(len(repr(merge.query('9781484206546', 'serial'))) < 20, True)  # NO METADATA



# -*- coding: utf-8 -*-
"""Base class to query a webservice and parse the result to py objects."""

import json
import logging
from time import sleep, time as timestamp

from . import webservice
from ._exceptions import DataNotFoundAtServiceError, ServiceIsDownError

UA = 'isbnlib (gzip)'
OUT_OF_SERVICE = 'Temporarily out of service'
BOOK_NOT_FOUND = 'No results match your search'
LOGGER = logging.getLogger(__name__)
THROTTLING = 1


class WEBQuery(object):

    """Base class to query a webservice and parse the result to py objects."""

    T = {'id': timestamp()}  # noqa

    def __init__(self, service_url, ua=UA):
        """Initialize & call webservice."""
        srv = service_url[8:20]
        last = WEBQuery.T[srv] if srv in WEBQuery.T else 0.0
        wait = 0 if timestamp() - last > THROTTLING else THROTTLING
        sleep(wait)
        self.url = service_url
        self.data = webservice.query(service_url, ua)
        WEBQuery.T[srv] = timestamp()

    def check_data(self, data_checker=None):  # pragma: no cover
        """Check the data & handle errors."""
        if data_checker:
            return data_checker(self.data)
        if self.data == '{}':
            LOGGER.warning('DataNotFoundAtServiceError for %s', self.url)
            raise DataNotFoundAtServiceError(self.url)
        if BOOK_NOT_FOUND in self.data:
            LOGGER.warning('DataNotFoundAtServiceError for %s', self.url)
            raise DataNotFoundAtServiceError(self.url)
        if OUT_OF_SERVICE in self.data:
            LOGGER.critical('ServiceIsDownError for %s', self.url)
            raise ServiceIsDownError(self.url)
        return True

    def parse_data(self, parser=json.loads):
        """Parse the data (default JSON -> PY)."""
        if parser is None:  # pragma: no cover
            return self.data
        return parser(self.data)   # <-- data is now unicode


def query(url, user_agent=UA, data_checker=None, parser=json.loads):
    """Put call and return the data from the web service."""
    wq = WEBQuery(url, user_agent)
    return wq.parse_data(parser) \
        if wq.check_data(data_checker) else None



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

import locale
import os
from nose.tools import assert_equals, assert_raises
from .._ext import ren
from ..dev._bouth23 import b2u3, u
from ..dev.helpers import File, cwdfiles

"""
nose tests
"""

WINDOWS = os.name == 'nt'
ENCODING = locale.getpreferredencoding()
if ENCODING != 'UTF-8':
    print("Your default locale encoding (%s) doesn't allow unicode filenames!" % ENCODING)
    print("=> Some tests could fail.")

TESTFILE_1 = './ç-deleteme.pdf' if WINDOWS else '/tmp/ç-deleteme.pdf'
TESTFILE_2 = './ç-deleteme-PLEASE.pdf' if WINDOWS else '/tmp/ç-deleteme-PLEASE.pdf'

F1 = '9780321534965.pdf'
F2 = '9781597499644.pdf'
F3 = '9781852330729.pdf'
F4 = '9787500117018.pdf'
F5 = '9789727576807.pdf'

F6 = 'Campos2011_Emergências obstétricas_9789727576807.pdf'
F7 = 'Knuth2008_The Art Of Computer Programming_9780321534965.pdf'
F8 = 'Man2001_Genetic Algorithms Concepts And Designs_9781852330729.pdf'
F9 = "O'Connor2012_Violent Python A Cookbook for Hackers, Forensic Analysts, Penetra_9781597499644.pdf"
F10 = '海明威2007_Lao ren yu hai_9787500117018.pdf'

F11 = 'myfile.pdf'

FISBN = [F1, F2, F3, F4, F5]
FFT = [F6, F7, F8, F9, F10]
FILES = FISBN + FFT + [F11]


def create_files(files):
    os.chdir(os.path.dirname(TESTFILE_1))
    for fn in files:
        try:
            with open(fn, 'w') as f:
                f.write(b2u3('ooo') + b2u3(fn))
        except UnicodeEncodeError:
            print("Your default locale (%s) doesn't allow non-ascii filenames!" % locale.CODESET)


def delete_files(fnpatt):
    os.chdir(os.path.dirname(TESTFILE_1))
    for fn in cwdfiles(fnpatt):
        os.remove(fn)


def setup_module():
    # create_files([u(TESTFILE_1), u(TESTFILE_2)])
    os.chdir(os.path.dirname(TESTFILE_1))
    create_files(FISBN+[F11])


def teardown_module():
    delete_files("*.pdf")


def test_ren():
    """Test 'high level' ren function."""
    ren(F1)
    assert_equals('Knuth2008_The Art Of Computer Programming_9780321534965.pdf' in cwdfiles("*.pdf"), True)
    # create_files([F5])
    # ren(F5)
    # assert_equals('Campos2011_Emergências obstétricas_9789727576807.pdf' in cwdfiles("*.pdf"), True)




# -*- coding: utf-8 -*-
"""Return a small description of the book."""

from json import loads
from textwrap import fill

from .dev.webservice import query as wsquery


UA = "isbnlib (gzip)"


def goo_desc(isbn):
    """Get description from Google Books api."""
    from .registry import metadata_cache  # <-- dynamic
    cache = metadata_cache
    if cache is not None:                     # pragma: no cover
        key = 'gdesc' + isbn
        try:
            if cache[key]:
                return cache[key]
            else:
                raise  # <-- IMPORTANT: usually the caches don't return error!
        except:
            pass
    url = "https://www.googleapis.com/books/v1/volumes?q=isbn+{isbn}"\
          "&fields=items/volumeInfo(description)"\
          "&maxResults=1".format(isbn=isbn)
    content = wsquery(url, user_agent=UA)
    try:
        content = loads(content)
        content = content['items'][0]['volumeInfo']['description']
        content = fill(content, width=75) if content else None
        if content and cache is not None:     # pragma: no cover
            cache[key] = content
        return content
    except KeyError:                          # pragma: no cover
        return



# -*- coding: utf-8 -*-

"""Read and write to a dict-like cache."""


from collections import MutableMapping


class IMCache(MutableMapping):

    """Read and write to a dict-like cache."""

    MAXLEN = 1000

    def __init__(self, maxlen=MAXLEN, *a, **k):
        self.filepath = 'IN MEMORY'
        self.maxlen = maxlen
        self.d = dict(*a, **k)
        while len(self) > maxlen:  # pragma: no cache
            self.popitem()

    def __iter__(self):
        return iter(self.d)

    def __len__(self):
        return len(self.d)

    def __getitem__(self, k):
        return self.d[k]

    def __setitem__(self, k, v):
        if k not in self and len(self) == self.maxlen:
            self.popitem()
        self.d[k] = v

    def __delitem__(self, k):
        del self.d[k]



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file


from .._infogroup import infogroup
from .._ext import info
from nose.tools import assert_equals, assert_raises


# nose tests
def test_infogroup():
    """Test 'infogroup' language/country function."""
    assert_equals(infogroup('9789727576807'), 'Portugal')
    assert_equals(infogroup('978-972-757-680-7'), 'Portugal')
    assert_equals(infogroup('7500117019'), "China, People's Republic")
    assert_equals(infogroup('7-5001-1701-9'), "China, People's Republic")
    assert_equals(infogroup('9524712946'), 'Finland')
    assert_equals(infogroup('0330284983'),
                  'English language')
    assert_equals(infogroup('3796519008'),
                  'German language')
    assert_raises(Exception, infogroup, '92xxxxxxxxxxx')
    assert_raises(Exception, infogroup, '')
    assert_equals(infogroup('9791090636071'), 'France')
    assert_equals(infogroup('9786131796364'), 'Mauritius')
    assert_equals(infogroup('9789992158104'), 'Qatar')

def test_ext_info():
    """Test 'info' language/country function."""
    assert_equals(info('9524712946'), 'Finland')
    assert_raises(Exception, info, '')



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""Format canonical in bibliographic formats."""

import re
import uuid
from string import Template

from ._helpers import last_first

bibtex = r"""@book{$ISBN,
     title = {$Title},
    author = {$AUTHORS},
      isbn = {$ISBN},
      year = {$Year},
 publisher = {$Publisher}
}"""

endnote = r"""%0 Book
%T $Title
%A $AUTHORS
%@ $ISBN
%D $Year
%I $Publisher """

refworks = r"""TY  - BOOK
T1  - $Title
A1  - $AUTHORS
SN  - $ISBN
Y1  - $Year
PB  - $Publisher
ER  - """

msword = r'''<b:Source xmlns:b="http://schemas.microsoft.com/office/'''\
         r'''word/2004/10/bibliography">
<b:Tag>$uid</b:Tag>
<b:SourceType>Book</b:SourceType>
<b:Author>
<b:NameList>$AUTHORS
</b:NameList>
</b:Author>
<b:Title>$Title</b:Title>
<b:Year>$Year</b:Year>
<b:City></b:City>
<b:Publisher>$Publisher</b:Publisher>
</b:Source>'''

json = r'''{"type": "book",
     "title": "$Title",
    "author": [$AUTHORS],
      "year": "$Year",
"identifier": [{"type": "ISBN", "id": "$ISBN"}],
 "publisher": "$Publisher"}'''

opf = r"""<?xml version='1.0' encoding='utf-8'?>
<package version="2.0" xmlns="http://www.idpf.org/2007/opf" unique-identifier="uuid_id">
  <metadata xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:opf="http://www.idpf.org/2007/opf">
    <dc:type>Book</dc:type>
    <dc:identifier opf:scheme="uuid" id="uuid_id">$uid</dc:identifier>
    <dc:identifier opf:scheme="ISBN" id="isbn_id">$ISBN</dc:identifier>
    <dc:title>$Title</dc:title>
    $AUTHORS
    <dc:publisher>$Publisher</dc:publisher>
    <dc:date>$Year</dc:date>
    <dc:contributor opf:file-as="isbnlib" opf:role="mdc">isbnlib [http://github.com/xlcnd/isbnlib]</dc:contributor>
  </metadata>
</package>"""

labels = r"""Type:      BOOK
Title:     $Title
Author:    $AUTHORS
ISBN:      $ISBN
Year:      $Year
Publisher: $Publisher"""

templates = {
             'labels': labels,
             'bibtex': bibtex,
             'endnote': endnote,
             'refworks': refworks,
             'msword': msword,
             'json': json,
             'opf': opf
             }

fmts = list(templates.keys())


def _gen_proc(name, canonical):
    if 'ISBN-13' in canonical:
        canonical['ISBN'] = canonical.pop('ISBN-13')
    tpl = templates[name]
    return Template(tpl).safe_substitute(canonical)


def _spec_proc(name, fmtrec, authors):
    """Fix the Authors records."""
    if name not in fmts:
        return
    if name == 'labels':
        AUTHORS = '\nAuthor:    '.join(authors)
    elif name == 'bibtex':
        AUTHORS = ' and '.join(authors)
    elif name == 'refworks':
        AUTHORS = '\nA1  - '.join(authors)
    elif name == 'endnote':
        AUTHORS = '\n%A '.join(authors)
    elif name == 'msword':
        fmtrec = fmtrec.replace('$uid', str(uuid.uuid4()))
        person = r"<b:Person><b:Last>$last</b:Last>"\
                 r"<b:First>$first</b:First></b:Person>"
        AUTHORS = '\n'.join(
            Template(person).safe_substitute(last_first(a))
            for a in authors)
    elif name == 'json':
        AUTHORS = ', '.join('{"name": "$"}'.replace("$", a)
                            for a in authors)
    elif name == 'opf':
        fmtrec = fmtrec.replace('$uid', str(uuid.uuid4()))
        creator = r'<dc:creator opf:file-as="$last, $first"'\
                  r' opf:role="aut">$first $last</dc:creator>'
        AUTHORS = '\n    '.join(
            Template(creator).safe_substitute(last_first(author))
            for author in authors)
    return re.sub(r'\$AUTHORS', AUTHORS, fmtrec)


def fmtbib(fmtname, canonical):
    """Return a canonical record in the selected format."""
    return _spec_proc(fmtname, _gen_proc(fmtname, canonical),
                      canonical['Authors'])



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

""" nose tests

"""

import locale
import os
from ..dev._files import File, cwdfiles
from nose.tools import assert_equals, assert_raises

WINDOWS = os.name == 'nt'
ENCODING = locale.getpreferredencoding()
if ENCODING == 'UTF-8':
    TESTFILE = './ç-deleteme.pdf' if WINDOWS else '/tmp/海明威-deleteme.pdf'
    NEW_BASENAME = 'ç-deleteme-PLEASE.pdf' if WINDOWS else '海明威-deleteme-PLEASE.pdf'
else:
    print("Your default locale encoding (%s) doesn't allow unicode filenames!" % ENCODING)
    TESTFILE = './deleteme.pdf'
    NEW_BASENAME = 'deleteme-PLEASE.pdf'

def setup_module():
    with open(TESTFILE, 'w') as f:
            f.write('ooo')
    os.chdir(os.path.dirname(TESTFILE))

def teardown_module():
    os.remove(os.path.join(os.path.dirname(TESTFILE), NEW_BASENAME))

def test_isfile():
    """Test if a path is a file."""
    f = File(TESTFILE)
    assert f.isfile(TESTFILE) == True

def test_exists():
    """Test if a path is a file or a directory."""
    f = File(TESTFILE)
    assert f.exists(TESTFILE) == True

def test_validate():
    """Test if a string is a valid filename for 'ren' command."""
    f = File(TESTFILE)
    assert f.validate('basename.pdf') == True
    assert f.validate('as/basename.pdf') == False
    assert f.validate('.basename.pdf') == True
    assert f.validate('.basename') == False
    assert f.validate('') == False


def test_mkwinsafe():
    """Test if a string is a valid basename in Windows."""
    f = File(TESTFILE)
    assert f.mkwinsafe('Açtr: ') == 'Açtr'
    assert f.mkwinsafe('as/tiõ') == 'astiõ'
    assert f.mkwinsafe('file ""name?') == 'file name'
    assert f.mkwinsafe('file ""name?', space='_') == 'file_name'
    assert f.mkwinsafe('file   name ') == 'file name'


def test_baserename():
    """Test the rename of a basename."""
    f = File(TESTFILE)
    assert f.baserename(NEW_BASENAME) == True
    assert f.baserename(NEW_BASENAME) == True


def test_cwdfiles():
    """Test the renaming of files in cwd."""
    assert (NEW_BASENAME in cwdfiles()) == True
    assert (NEW_BASENAME in cwdfiles('*.pdf')) == True
    assert (NEW_BASENAME in cwdfiles('*.txt')) == False



# -*- coding: utf-8 -*-
"""Return editions for a given ISBN."""

import logging

from ._core import EAN13
from ._exceptions import NotRecognizedServiceError, NotValidISBNError
from ._thinged import query as ted
from ._wcated import query as wed

PROVIDERS = ('any', 'merge', 'thingl', 'wcat')
TRUEPROVIDERS = ('wcat', 'thingl')  # <-- by priority
LOGGER = logging.getLogger(__name__)


def fake_provider_any(isbn):
    """Fake provider 'any' service."""
    providers = {'wcat': wed, 'thingl': ted}
    for provider in TRUEPROVIDERS:
        data = []
        try:
            data = providers[provider](isbn)
            if data:
                return data
            continue       # pragma: no cover
        except:            # pragma: no cover
            continue
    return data            # pragma: no cover


def fake_provider_merge(isbn):
    """Fake provider 'merge' service."""
    data = []
    try:                   # pragma: no cover
        wdata = wed(isbn)
        if wdata:
            tdata = ted(isbn)
            if tdata:
                data = list(set(wdata + tdata))
                return data
        raise
    except:                # pragma: no cover
        return data


def editions(isbn, service='wcat'):
    """Return the list of ISBNs of editions related with this ISBN."""
    isbn = EAN13(isbn)
    if not isbn:
        LOGGER.critical('%s is not a valid ISBN', isbn)
        raise NotValidISBNError(isbn)

    if service == 'any':
        return fake_provider_any(isbn)

    if service == 'merge':
        return fake_provider_merge(isbn)

    if service not in PROVIDERS:
        LOGGER.critical('%s is not a recognized editions provider', service)
        raise NotRecognizedServiceError(service)
    return wed(isbn) if service == 'wcat' else ted(isbn)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""
nose tests
"""

from nose.tools import assert_equals
from ..dev._bouth23 import u
from ..dev._fmt import fmtbib


canonical = {
             'ISBN-13': u('9780123456789'),
             'Title': u('A book about nothing'),
             'Publisher': u('No Paper Press'),
             'Year': u('2000'),
             'Language': u('en'),
             'Authors': [u('John Smith'), u('José Silva')]
             }

def test_fmtbib():
    """Test the formating into several bibliographic formats."""
    assert_equals(len(fmtbib("bibtex", canonical)), 182)
    assert_equals(len(fmtbib("labels", canonical)), 158)
    assert_equals(len(fmtbib("endnote", canonical)), 103)
    assert_equals(len(fmtbib("msword", canonical)), 485)
    assert_equals(len(fmtbib("json", canonical)), 229)
    assert_equals(len(fmtbib("refworks", canonical)), 130)
    assert_equals(len(fmtbib("opf", canonical)), 861)



# -*- coding: utf-8 -*-
"""Get the Language/Country of an ISBN."""

import logging

from ._core import EAN13
from ._data.data4info import d, identifiers
from ._exceptions import NotValidISBNError

LOGGER = logging.getLogger(__name__)


def infogroup(isbn):
    """Get the Language/Country of this ISBN."""
    # if isbn is not a valid ISBN this def can give a wrong result!
    # => clean and validate
    isbn = EAN13(isbn)
    if not isbn:
        LOGGER.critical('%s is not a valid ISBN', isbn)
        raise NotValidISBNError(isbn)
    # put isbn in the form 978-...
    prefix = isbn[0:3] + '-'
    isbn = prefix + isbn[3:]
    dtxt = d
    idents = identifiers
    ixi, ixf = 4, 5
    for ident in idents:
        iid = prefix + isbn[ixi:ixf]
        ixf += 1
        # stop if identifier is found
        if iid in ident:
            return dtxt[iid]



# -*- coding: utf-8 -*-

"""Private helper functions."""

import re
import sys
from hashlib import md5

from ._bouth23 import b, s


def fake_isbn(title, author='unkown', publisher='unkown', sid=1):
    """Produce a fake ISBN from the (title, author, publisher) of the book."""
    key = "%s %s %s" % (title, author, publisher)
    # normalize
    regex1 = re.compile(r'\?|,|\.|!|\:|;', re.I | re.M | re.S)
    regex2 = re.compile(r'\s\s+', re.I | re.M | re.S)
    key = regex1.sub(' ', key)
    key = regex2.sub(' ', key).strip().lower()
    # hash
    return (str(sid) + str(int(md5(b(key)).hexdigest()[:10], 16)))[:13]


def in_virtual():       # pragma: no cover
    """Detect if program is running inside a python virtual environment."""
    return True if hasattr(sys, 'real_prefix') else False


def normalize_space(item):
    """Normalize white space.

    Strips leading and trailing white space and replaces sequences of
    white space characters with a single space.
    """
    item = re.sub(r'\s\s+', ' ', item)
    return item.strip()


def titlecase(st):
    """Format string in title case (for ascii).

    Only changes the first character of each word.
    """
    try:
        st.encode('ascii')
        return re.sub(r"[A-Za-z]+('[A-Za-z]+)?",
                      lambda m: m.group(0)[0].upper() + m.group(0)[1:], st)
    except UnicodeDecodeError:  # pragma: no cover
        return st


def last_first(author):
    """Parse an author name into last (name) and first."""
    if ',' in author:
        tokens = author.split(',')
        last = tokens[0].strip()
        first = ' '.join(tokens[1:]).strip().replace('  ', ', ')
    else:
        tokens = author.split(' ')
        last = tokens[-1].strip()
        first = ' '.join(tokens[:-1]).strip()
    return {'last': last, 'first': first}


def unicode_to_utf8tex(utex, filtre=()):
    """Replace unicode entities with tex entitites and returns utf8 bytes."""
    from .._data.data4tex import unicode_to_tex
    btex = utex.encode('utf-8')
    table = dict((k.encode('utf-8'), v) for k, v in unicode_to_tex.items()
                 if v not in filtre)
    regex = re.compile(b('|'.join(re.escape(s(k)) for k in table)))
    return regex.sub(lambda m: table[m.group(0)], btex)


def cutoff_tokens(tokens, cutoff):
    """Keep only the tokens with total length <= cutoff."""
    ltokens = [len(t) for t in tokens]
    length = 0
    stokens = []
    for token, l in zip(tokens, ltokens):
        if length + l <= cutoff:
            length = length + l
            stokens.append(token)
        else:
            break
    return stokens


def parse_placeholders(pattern):
    """Return a list of placeholders in a pattern."""
    regex = re.compile(r'({[^}]*})')
    return regex.findall(pattern)



# -*- coding: utf-8 -*-
"""Provide metadata by merging metadata from other providers."""

from . import config
from ._goob import query as qgoob
from ._wcat import query as qwcat
from .dev import Metadata, vias


def query(isbn, processor=None):
    """Query function for the 'merge provider' (waterfall model)."""
    if not processor:
        processor = config.options.get('VIAS_MERGE', processor).lower()
        if not processor:     # pragma: no cover
            processor = 'serial'

    named_tasks = (('wcat', qwcat), ('goob', qgoob))
    if processor == 'parallel':
        results = vias.parallel(named_tasks, isbn)
    elif processor == 'serial':
        results = vias.serial(named_tasks, isbn)
    elif processor == 'multi':
        results = vias.multi(named_tasks, isbn)

    rw = results.get('wcat', None)
    rg = results.get('goob', None)

    if not rw and not rg:     # pragma: no cover
        return None

    md = Metadata(rw) if rw else None

    if md and rg:
        # Overwrite with Authors and Language from Google
        md.merge(rg, overwrite=('Authors'))
        return md.value
    if not md and rg:         # pragma: no cover
        md = Metadata(rg)
        return md.value
    return md.value if not rg and rw else None  # pragma: no cover



# -*- coding: utf-8 -*-
"""Use Google to get an ISBN from words from title and author's name."""


import logging

from ._core import get_canonical_isbn, get_isbnlike
from .dev import webservice

LOGGER = logging.getLogger(__name__)


def goos(words):
    """Use Google to get an ISBN from words from title and author's name."""
    service_url = "http://www.google.com/search?q={words}+ISBN"
    search_url = service_url.format(words=words.replace(' ', '+'))
    user_agent = 'w3m/0.5.2'

    content = webservice.query(search_url, user_agent)
    isbns = get_isbnlike(content)

    for item in isbns:
        isbn = get_canonical_isbn(item, output='isbn13')
        if isbn:
            break
    if not isbns or not isbn:    # pragma: no cover
        LOGGER.debug('No ISBN found for %s', words)
        return
    return isbn



# -*- coding: utf-8 -*-
"""Query web services."""

import gzip
import logging

from ._bouth23 import bstream, s
from ._exceptions import ISBNLibHTTPError, ISBNLibURLError

try:                     # pragma: no cover
    from urllib.parse import urlencode
    from urllib.request import Request, urlopen
    from urllib.error import HTTPError, URLError
except ImportError:      # pragma: no cover
    from urllib import urlencode
    from urllib2 import Request, urlopen, HTTPError, URLError

UA = 'isbnlib (gzip)'
LOGGER = logging.getLogger(__name__)


class WEBService(object):

    """Class to query web services."""

    def __init__(self, url, user_agent=UA, values=None, appheaders=None):
        """Initialize main properties."""
        self._url = url
        # headers to accept gzipped content
        headers = {'Accept-Encoding': 'gzip', 'User-Agent': user_agent}
        # add more user provided headers
        if appheaders:  # pragma: no cover
            headers.update(appheaders)
        # if 'data' it does a PUT request (data must be urlencoded)
        data = urlencode(values) if values else None
        self._request = Request(url, data, headers=headers)
        self.response = None

    def _response(self):
        """Check errors on response."""
        try:
            self.response = urlopen(self._request)
            LOGGER.debug('Request headers:\n%s', self._request.header_items())
        except HTTPError as e:  # pragma: no cover
            LOGGER.critical('ISBNLibHTTPError for %s with code %s [%s]',
                            self._url, e.code, e.msg)
            if e.code in (401, 403, 429):
                raise ISBNLibHTTPError('%s Are you making many requests?'
                                       % e.code)
            if e.code in (502, 504):
                raise ISBNLibHTTPError('%s Service temporarily unavailable!'
                                       % e.code)
            raise ISBNLibHTTPError('(%s) %s' % (e.code, e.msg))
        except URLError as e:   # pragma: no cover
            LOGGER.critical('ISBNLibURLError for %s with reason %s',
                            self._url, e.reason)
            raise ISBNLibURLError(e.reason)

    def data(self):
        """Return the uncompressed data."""
        self._response()
        LOGGER.debug('Response headers:\n%s', self.response.info())
        if self.response.info().get('Content-Encoding') == 'gzip':
            buf = bstream(self.response.read())
            f = gzip.GzipFile(fileobj=buf)
            data = f.read()
        else:                   # pragma: no cover
            data = self.response.read()
        return s(data)


def query(url, user_agent=UA, values=None, appheaders=None):
    """Query to a web service."""
    service = WEBService(url, user_agent=user_agent, values=values,
                         appheaders=appheaders)
    data = service.data()
    LOGGER.debug('Raw data from service:\n%s', data)
    return data



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""
nose tests
"""

from ..dev._helpers import last_first, cutoff_tokens, parse_placeholders, fake_isbn
from nose.tools import assert_equals


def test_last_first():
    """Test the parsing of author's name into (Surname, First Name)."""
    assert_equals(last_first("Surname, First Name"), {"last": "Surname", "first": "First Name"})
    assert_equals(last_first("First Name Surname"), {"last": "Surname", "first": "First Name"})
    assert_equals(last_first("Surname1, First1 and Sur2, First2"), {"last": "Surname1", "first": "First1 and Sur2, First2"})


def test_cutoff_tokens():
    """Test the 'cutoff_tokens' function."""
    assert_equals(cutoff_tokens(['1', '23', '456'], 3), ['1', '23'])


def test_parse_placeholders():
    """Test the parsing of placeholders."""
    assert_equals(parse_placeholders('{isbn}_akaj_{name}'), ['{isbn}', '{name}'])


def test_fake_isbn():
    """Test the 'fake_isbn' function."""
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; '), '1111006407537')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ', author=''), '1108449680873')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ', author=' '), '1108449680873')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ', author='', publisher=''), '1181593982422')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ', author='a', publisher='K'), '1895031085488')
    assert_equals(fake_isbn(' Hello?? Wer,  ! ksDf:  asdf. ; ', author='A', publisher='k'), '1895031085488')



# -*- coding: utf-8 -*-
"""Config file for isbnlib."""

# --> Import only external modules! <--


import socket

# Timeouts
SOCKETS_TIMEOUT = 12    # seconds
THREADS_TIMEOUT = 11    # seconds


def setsocketstimeout(seconds):
    """Set the value of SOCKETS_TIMEOUT (in seconds)."""
    global SOCKETS_TIMEOUT
    SOCKETS_TIMEOUT = seconds
    return socket.setdefaulttimeout(SOCKETS_TIMEOUT)

# socket timeout is not exposed at urllib2 level so I had to import the
# module and set a default value for all the sockets (timeout in seconds)
# however this should be done at top level due to strong side effects...
setsocketstimeout(SOCKETS_TIMEOUT)


# THREADS_TIMEOUT is a parameter used downstream by Thread calls (see vias.py)
def setthreadstimeout(seconds):   # pragma: no cover
    """Set the value of THREADS_TIMEOUT (in seconds)."""
    global THREADS_TIMEOUT
    THREADS_TIMEOUT = seconds


# API keys
apikeys = {}


def add_apikey(service, apikey):  # pragma: no cover
    """Add API keys.

    add_apikey('isbndb', 'JuHytr6') [is fake!]
    """
    apikeys[service] = apikey


# Generic Options
options = {'VIAS_MERGE': 'serial'}


def set_option(option, value):    # pragma: no cover
    """Set the value for option."""
    options[option] = value



# -*- coding: utf-8 -*-

"""Helper module to work with files."""


import fnmatch
import logging
import os
import re
from stat import S_IRGRP, S_IROTH, S_IRUSR, S_IWGRP, S_IWOTH, S_IWUSR

from ._exceptions import FileNotFoundError

MAXLEN = 120
ILEGAL = r'<>:"/\|?*'
LOGGER = logging.getLogger(__name__)
MODE666 = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH


class File(object):

    """Easy manipulation of files in the SAME directory."""

    def __init__(self, fp):
        """Set and validate the basic properties."""
        if not self.isfile(fp):
            raise FileNotFoundError(fp)
        self.path = os.path.dirname(fp) or os.getcwd()
        self.basename = os.path.basename(fp)
        self.name, self.ext = os.path.splitext(self.basename)
        self.writable = os.access(fp, os.W_OK)

    def siblings(self):
        """Collect files and directories in the same directory."""
        return [f for f in os.listdir(self.path) if f != self.basename]

    @staticmethod
    def isfile(path):
        """Check if a given path is a file."""
        return os.path.isfile(path)

    @staticmethod
    def exists(path):
        """Check if a given path is a file or a directory."""
        return os.path.exists(path)

    @staticmethod
    def mkwinsafe(name, space=' '):
        """Delete most common characters not allowed in Windows filenames."""
        space = space if space not in ILEGAL else ' '
        name = ''.join(c for c in name if c not in ILEGAL)\
               .replace(' ', space).strip()
        name = re.sub(r'\s\s+', ' ', name) if space == ' ' else name
        return name[:MAXLEN]

    @staticmethod
    def validate(basename):
        """Check for a proper basename."""
        if basename != os.path.basename(basename):
            LOGGER.critical("This (%s) is not a basename!", basename)
            return False
        name, ext = os.path.splitext(basename)
        if len(name) == 0:
            LOGGER.critical("Not a valid name (lenght 0)!")
            return False
        if len(ext) == 0:
            LOGGER.critical("Not a valid extension (lenght 0)!")
            return False
        return True

    def baserename(self, new_basename):
        """Rename the file to a 'safe' basename."""
        if not self.validate(new_basename):
            return False
        name, ext = os.path.splitext(new_basename)
        name = self.mkwinsafe(name)
        new_basename = name + ext
        if new_basename == self.basename:
            return True
        if new_basename not in self.siblings():
            try:
                os.rename(self.basename, new_basename)
            except OSError as err:
                LOGGER.critical("%s", err.message)
                return False
            self.basename = new_basename
            self.name = name
            self.ext = ext
            return True
        else:
            LOGGER.info("The file (%s) already exist in the directory!",
                        new_basename)
            return True

    @staticmethod
    def uxchmod(fp, mode=MODE666):
        """Change the mode of the file (default is 0666)."""
        return os.chmod(fp, mode)


def cwdfiles(pattern='*'):
    """List the files in current directory that match a given pattern."""
    return fnmatch.filter(os.listdir('.'), pattern)



# -*- coding: utf-8 -*-
"""Query the Google Books (JSON API v1) service for metadata."""

import logging

from .dev import stdmeta
from .dev._bouth23 import u
from .dev._exceptions import (ISBNNotConsistentError,
                              NoDataForSelectorError, RecordMappingError)
from .dev.webquery import query as wquery

UA = 'isbnlib (gzip)'
SERVICE_URL = 'https://www.googleapis.com/books/v1/volumes?q=isbn+{isbn}'\
    '&fields=items/volumeInfo(title,authors,publisher,publishedDate,language,'\
    'industryIdentifiers)&maxResults=1'
LOGGER = logging.getLogger(__name__)


def _mapper(isbn, records):
    """Mapp: canonical <- records."""
    # canonical: ISBN-13, Title, Authors, Publisher, Year, Language
    try:
        canonical = {}
        canonical['ISBN-13'] = u(isbn)
        canonical['Title'] = records.get('title', u('')).replace(' :', ':')
        canonical['Authors'] = records.get('authors', [u('')])
        canonical['Publisher'] = records.get('publisher', u(''))
        if 'publishedDate' in records \
           and len(records['publishedDate']) >= 4:
            canonical['Year'] = records['publishedDate'][0:4]
        else:         # pragma: no cover
            canonical['Year'] = u('')
        canonical['Language'] = records.get('language', u(''))
    except:           # pragma: no cover
        LOGGER.debug("RecordMappingError for %s with data %s", isbn, records)
        raise RecordMappingError(isbn)
    # call stdmeta for extra cleanning and validation
    return stdmeta(canonical)


def _records(isbn, data):
    """Classify (canonically) the parsed data."""
    # put the selected data in records
    try:
        recs = data['items'][0]['volumeInfo']
    except:  # pragma: no cover
        LOGGER.debug('NoDataForSelectorError for %s', isbn)
        raise NoDataForSelectorError(isbn)
    # consistency check (isbn request = isbn response)
    if recs:
        ids = recs.get('industryIdentifiers', '')
        if isbn not in repr(ids):   # pragma: no cover
            LOGGER.debug('ISBNNotConsistentError for %s (%s)', isbn, repr(ids))
            raise ISBNNotConsistentError("%s not in %s" % (isbn, repr(ids)))
    # map canonical <- records
    return _mapper(isbn, recs)


def query(isbn):
    """Query the Google Books (JSON API v1) service for metadata."""
    data = wquery(SERVICE_URL.format(isbn=isbn), user_agent=UA)
    return _records(isbn, data)



# -*- coding: utf-8 -*-
"""Handle metadata objects."""

import logging

from ._bouth23 import type3str, u
from ._exceptions import NotValidMetadataError
from ._helpers import normalize_space, titlecase

# For now you cannot add custom fields!
FIELDS = ('ISBN-13', 'Title', 'Authors', 'Publisher', 'Year', 'Language')

LOGGER = logging.getLogger(__name__)


class Metadata(object):

    """Class for metadata objects."""

    def __init__(self, record=None):
        """Initialize attributes."""
        self._content = None
        self._set_empty()
        if record:
            self._content.update((k, v) for k, v in list(record.items()))
            if not self._validate():
                self._set_empty()
                LOGGER.debug(record)
                raise NotValidMetadataError()
            self.clean()

    @staticmethod
    def fields():  # pragma: no cover
        """Return a list of value's fields."""
        return list(FIELDS)

    def clean(self, broom=normalize_space, exclude=()):
        """Clean fields of value."""
        self._content.update((k, broom(v)) for k, v
                             in list(self._content.items())
                             if k != 'Authors' and k not in exclude)
        if 'Authors' not in exclude:
            self._content['Authors'] = [broom(i) for i in
                                        self._content['Authors']]
        self._content['Title'] = self._content['Title'].strip(',.:;-_ ')
        if self._content['Language'].lower() in ('en', 'eng', 'english'):
            self._content['Title'] = titlecase(self._content['Title'])

    @property
    def value(self):
        """Get value."""
        return self._content

    @value.setter
    def value(self, record):  # pragma: no cover
        """Set value."""
        self._content.update((k, v) for k, v in list(record.items()))
        if not self._validate():
            self._set_empty()
            LOGGER.debug(record)
            raise NotValidMetadataError()
        self.clean()

    @value.deleter
    def value(self):  # pragma: no cover
        """Delete value."""
        self._set_empty()

    def merge(self, record, overwrite=(),
              overrule=lambda x: x == u('') or x == [u('')]):
        """Merge the record with value."""
        # by default do nothing
        self._content.update((k, v) for k, v in list(record.items())
                             if k in overwrite and not overrule(v) or
                             self._content[k] == u(''))
        if not self._validate():  # pragma: no cover
            self._set_empty()
            LOGGER.debug(record)
            raise NotValidMetadataError()
        self.clean()

    def _validate(self):
        """Validate value."""
        # 'minimal' check
        for k in self._content:
            if not type(self._content[k]) is type3str():
                if k != 'Authors':
                    return False
        if not type(self._content['Authors']) is list:
            return False
        return True

    def _set_empty(self):
        """Set an empty value record."""
        self._content = dict.fromkeys(list(FIELDS), u(''))
        self._content['Authors'] = [u('')]


def stdmeta(records):
    """Function API to the class."""
    dt = Metadata(records)
    return dt.value



# -*- coding: utf-8 -*-
# flake8:noqa
# pylint: skip-file

"""Help code to run in py2 and py3."""


import sys

if sys.version < '3':

    def s(x):
        return x

    def b(x):
        return x

    def u(x):
        try:
            return unicode(x, "utf-8")
        except TypeError:
            return x

    def b2u3(x):
        return x.encode("utf-8")

    def type3str():
        return type(u'')

    def bstream(x):
        from StringIO import StringIO
        return StringIO(x)
else:

    def s(x):
        return x.decode("utf-8", 'ignore')

    def b(x):
        return x.encode("utf-8")

    def u(x):
        return x

    def b2u3(x):
        return x

    def type3str():
        return type('')

    def bstream(x):
        from io import BytesIO
        return BytesIO(x)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file
""" nose tests

"""

from nose.tools import assert_equals, assert_raises
from ..dev import Metadata, stdmeta
from ..dev._bouth23 import u


def test_stdmeta():
    """Test the transformation of raw records into standard metadata."""
    # test stdmeta from data
    r={
       'ISBN-13': u('9780123456789 '),
       'Title': u('Bla. Bla /Title .'),
       'Publisher': u(''),
       'Year': u('2000'),
       'Language': u('en'),
       'Authors': [u('author1. mba'), u('author2   ')]
       }
    R={
       'ISBN-13': u('9780123456789'),
       'Title': u('Bla. Bla /Title'),
       'Publisher': u(''),
       'Year': u('2000'),
       'Language': u('en'),
       'Authors': [u('author1. mba'), u('author2')]
       }
    A={
       'ISBN-13': u('9780123456789 '),
       'Title': b'Bla. Bla /Title .',
       'Publisher': u(''),
       'Year': b'2000',
       'Language': u('en'),
       'Authors': [u('author1. mba'), u('author2   ')]
       }
    B={
       'ISBN-13': u('9780123456789'),
       'Title': u('Bla. Bla /Title .'),
       'Publisher': u(''),
       'Year': u('2000'),
       'Language': u('en'),
       'Authors': u('author1')
       }
    assert_equals(stdmeta(r), R)
    assert_equals(stdmeta(R), R)
    assert_raises(Exception, stdmeta, A)
    assert_raises(Exception, stdmeta, B)

def test_metaclass():
    """Test the creation of a Metadata class from raw records."""
    R={
       'ISBN-13': u('9780123456789'),
       'Title': u('Bla. Bla /Title'),
       'Publisher': u(''),
       'Year': u('2000'),
       'Language': u('en'),
       'Authors': [u('author1. mba'), u('author2')]
       }
    dt = Metadata(R)
    assert_equals(dt.value, R)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""
nose tests
"""

from .. import doi2tex
from nose.tools import assert_equals


def test_doi2tex():
    """Test the doi2tex service."""
    assert_equals(len(repr(doi2tex('10.2139/ssrn.2411669'))) > 50, True)



# -*- coding: utf-8 -*-
"""Query the worldcat.org xID service for related ISBNs."""

import logging
from ast import literal_eval

from .dev._exceptions import DataWrongShapeError, NoDataForSelectorError
from .dev.webquery import query as wquery

LOGGER = logging.getLogger(__name__)
UA = 'isbnlib (gzip)'
SERVICE_URL = 'http://xisbn.worldcat.org/webservices/xid/isbn/{isbn}?'\
              'method=getEditions&format=python'


def _editions(isbn, data):
    """Return the records from the parsed response."""
    # check status
    try:
        status = data['stat']
        if status != 'ok':
            raise  # pragma: no cover
    except:        # pragma: no cover
        LOGGER.debug('DataWrongShapeError for %s with status %s',
                     isbn, status)
        raise DataWrongShapeError("status: '%s' for isbn %s" % (status, isbn))
    # put the selected data in records
    try:
        recs = [ib['isbn'][0] for ib in data['list']]
    except:        # pragma: no cover
        LOGGER.debug('NoDataForSelectorError for %s', isbn)
        raise NoDataForSelectorError(isbn)
    return recs


def query(isbn):
    """Query the worldcat.org service for related ISBNs."""
    data = wquery(SERVICE_URL.format(isbn=isbn),
                  user_agent=UA,
                  parser=literal_eval)
    return _editions(isbn, data)



# -*- coding: utf-8 -*-
"""Query the openlibrary.org service for metadata."""

import logging
import re

from .dev import stdmeta
from .dev._bouth23 import u
from .dev._exceptions import NoDataForSelectorError, RecordMappingError
from .dev.webquery import query as wquery

UA = 'isbnlib (gzip)'
SERVICE_URL = 'http://openlibrary.org/api/books?bibkeys='\
    'ISBN:{isbn}&format=json&jscmd=data'
LOGGER = logging.getLogger(__name__)


def _mapper(isbn, records):
    """Map canonical <- records."""
    # canonical:
    # -> ISBN-13, Title, Authors, Publisher, Year, Language
    try:
        # mapping: canonical <- records
        canonical = {}
        canonical['ISBN-13'] = u(isbn)
        canonical['Title'] = records.get('title', u('')).replace(' :', ':')
        canonical['Authors'] = [a['name'] for a in
                                records.get('authors', ({'name': u('')},))]
        canonical['Publisher'] = records.get('publishers',
                                             [{'name': u('')}, ])[0]['name']
        canonical['Year'] = u('')
        strdate = records.get('publish_date', u(''))
        if strdate:        # pragma: no cover
            match = re.search(r'\d{4}', strdate)
            if match:
                canonical['Year'] = match.group(0)
    except:                # pragma: no cover
        LOGGER.debug("RecordMappingError for %s with data %s", isbn, records)
        raise RecordMappingError(isbn)
    # call stdmeta for extra cleanning and validation
    return stdmeta(canonical)


def _records(isbn, data):
    """Classify (canonically) the parsed data."""
    try:
        # put the selected data in records
        records = data['ISBN:%s' % isbn]
    except:   # pragma: no cover
        LOGGER.debug('NoDataForSelectorError for %s', isbn)
        raise NoDataForSelectorError(isbn)

    # map canonical <- records
    return _mapper(isbn, records)


def query(isbn):
    """Query the openlibrary.org service for metadata."""
    data = wquery(SERVICE_URL.format(isbn=isbn), user_agent=UA)
    return _records(isbn, data)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""nose tests for metadata."""

from random import randrange
from .._metadata import query
from .._ext import meta
from nose.tools import assert_equals, assert_raises


def test_query():
    """Test the query of metadata with 'low level' queries."""
    # test query from metadata
    assert_raises(Exception, query, '9781849692341', 'goog')
    assert_raises(Exception, query, '9781849692342', 'goob')
    assert_raises(Exception, query, '9781849692341', 'wcat')
    assert_equals(len(repr(query('9780321534965', 'wcat'))) > 150, True)
    assert_equals(len(repr(query('9780321534965'))) > 150, True)
    assert_equals(len(repr(query('9780321534965', 'merge'))) > 150, True)
    assert_equals(len(repr(query('9780321534965', 'goob'))) > 150, True)
    assert_equals(len(repr(query('9789934015960'))) > 150, True)
    assert_equals(len(repr(query(u'9781118241257'))) > 149, True)
    assert_raises(Exception, query, '9780000000', 'wcat', None)
    assert_raises(Exception, query, randrange(0, 1000000), 'wcat')


def test_ext_meta():
    """Test the query of metadata with 'high level' meta function."""
    # test meta from core
    assert_equals(len(repr(meta('9780321534965', 'wcat'))) > 150, True)
    assert_equals(len(repr(meta('9780321534965', 'merge'))) > 150, True)
    assert_equals(len(repr(meta('9780321534965'))) > 150, True)
    assert_raises(Exception, meta, '9780000000', 'wcat', None)
    assert_raises(Exception, meta, randrange(0, 1000000), 'wcat')
    assert_raises(Exception, meta, '9781849692342', 'goob', None)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

from nose.tools import assert_equals, raises

from .._exceptions import NotRecognizedServiceError, NotValidISBNError
from .._ext import editions


# nose tests


def test_editions1():
    """Test the 'editions' service."""
    assert_equals(len(editions('9780156001311', service='wcat')) > 19, True)
    assert_equals(len(editions('9780151446476', service='thingl')) > 19, True)
    assert_equals(len(editions('9780151446476', service='any')) > 19, True)
    assert_equals(len(editions('9780151446476', service='merge')) > 19, True)


@raises(NotValidISBNError)
def test_editions2():
    """Test the 'editions' service error detection (NotValidISBNError)."""
    editions('978')


@raises(NotRecognizedServiceError)
def test_editions3():
    """Test the 'editions' service error detection (NotRecognizedServiceError)."""
    editions('9780156001311', service='xxx')



# -*- coding: utf-8 -*-
# flake8: noqa
# isort:skip_file

"""Define isbnlib API and set lib environment."""

import logging as _logging

from ._exceptions import (quiet_errors, ISBNLibException,
                          NotRecognizedServiceError,
                          NotValidISBNError,
                          PluginNotLoadedError)

# config _logging for lib (NullHandler not available for py26)
try:
    _nh = _logging.NullHandler()
except:              # pragma: no cover
    class NullHandler(_logging.Handler):
        def emit(self, record):
            pass
    _nh = NullHandler()
_logging.getLogger('isbnlib').addHandler(_nh)

# configuration
from . import config                                # <-- first import

# main modules
from ._core import (is_isbn10, is_isbn13, to_isbn10, to_isbn13, clean,
                    canonical, notisbn, get_isbnlike, get_canonical_isbn,
                    EAN13)
from ._ext import (cover, desc, mask, meta, info, editions, isbn_from_words,
                   doi, ren)
from ._goom import query as goom
from ._doitotex import doi2tex

# Ranges Database date
from ._data.data4mask import RDDATE

# alias
ISBN13 = EAN13
ean13 = EAN13

# dunders

__all__ = ('is_isbn10', 'is_isbn13', 'clean', 'mask', 'info', 'meta',
           'to_isbn10', 'to_isbn13', 'get_isbnlike', 'notisbn',
           'ean13', 'EAN13', 'cover', 'desc',
           'canonical', 'get_canonical_isbn', 'editions', 'isbn_from_words',
           'quiet_errors', 'config', '__version__', '__support__',
           'doi', 'ren', 'ISBN13', 'ISBNLibException',
           'NotRecognizedServiceError', 'NotValidISBNError',
           'PluginNotLoadedError', 'goom', 'doi2tex', 'RDDATE')

__version__ = '3.5.6'                               # <-- literal IDs
__support__ = 'py26, py27, py33, py34, pypy'        # <-- literal IDs



# -*- coding: utf-8 -*-
"""Extra methods."""

import re

from ._core import EAN13
from ._cover import cover as gcover
from ._desc import goo_desc
from ._editions import editions as eds
from ._gwords import goos
from ._infogroup import infogroup
from ._metadata import query
from ._msk import msk
from .dev._bouth23 import b2u3, u
from .dev.helpers import File, cutoff_tokens, last_first


def mask(isbn, separator='-'):
    """`Mask` a canonical ISBN."""
    return msk(isbn, separator)


def meta(isbn, service='default', cache='default'):
    """Get metadata from worldcat.org ('wcat'), Google Books ('goob') , ..."""
    service = service if service else 'default'
    return query(isbn, service, cache) if isbn else None


def info(isbn):
    """Get language or country assigned to this ISBN."""
    return infogroup(isbn)


def editions(isbn, service='wcat'):
    """Return the list of ISBNs of editions related with this ISBN.

    'service' can have the values 'any', 'merge', 'thingl' and 'wcat' (default)
    """
    return eds(isbn, service)


def isbn_from_words(words):
    """Return the most probable ISBN from a list of words."""
    return goos(words)


def doi(isbn):
    """Return a DOI's ISBN-A from a ISBN-13."""
    return "10.%s.%s%s/%s%s" % \
           tuple(msk(EAN13(isbn), '-').split('-'))


def ren(fp):
    """Rename a file using metadata from an ISBN in his filename."""
    cfp = File(fp)
    isbn = EAN13(cfp.name)
    if not isbn:                               # pragma: no cover
        return
    data = meta(isbn)
    author = data.get('Authors', u('UNKNOWN'))
    if author != u('UNKNOWN'):                 # pragma: no cover
        author = last_first(author[0])['last']
    year = data.get('Year', u('UNKNOWN'))
    maxlen = 98 - (20 + len(author) + len(year))
    title = data.get('Title', u('UNKNOWN'))
    if title != u('UNKNOWN'):
        regex1 = re.compile(r'[.,_!?/\\]')
        regex2 = re.compile('\s\s+')
        title = regex1.sub(' ', title)
        title = regex2.sub(' ', title)
        title = title.strip()
    if title == u('UNKNOWN') or not title:     # pragma: no cover
        return
    if ' ' in title:                           # pragma: no cover
        tokens = title.split(' ')
        stitle = cutoff_tokens(tokens, maxlen)
        title = ' '.join(stitle)
    isbn13 = data.get('ISBN-13', u('UNKNOWN'))
    new_name = "%s%s_%s_%s" % (author, year, title, isbn13)
    return cfp.baserename(b2u3(new_name + cfp.ext))


def cover(isbn, size=2, mode='prt'):
    """Download the cover of the ISBN."""
    isbn = EAN13(isbn)
    return gcover(isbn, size=size, mode=mode) if isbn else None


def desc(isbn):
    """Return a descripion of the ISBN."""
    isbn = EAN13(isbn)
    return goo_desc(isbn) if isbn else None



# -*- coding: utf-8 -*-

"""Return metadata, of a given DOI, formated as BibTeX."""

import logging

from .dev.webservice import query

LOGGER = logging.getLogger(__name__)


URL = 'http://dx.doi.org/{DOI}'
UA = 'isbnlib (gzip)'


def doi2tex(doi):
    """Get the bibtex ref for doi."""
    data = query(URL.format(DOI=doi),
                 user_agent=UA,
                 appheaders={'Accept': 'application/x-bibtex; charset=utf-8'}
                 )
    if not data:  # pragma: no cover
        LOGGER.warning('no data return for doi: %s', doi)
        return
    return data



# -*- coding: utf-8 -*-
"""Query the ThingISBN api (Library Thing) for related ISBNs."""

import logging
from xml.dom.minidom import parseString

from ._core import EAN13
from .dev.webquery import query as wquery

LOGGER = logging.getLogger(__name__)
UA = 'isbnlib (gzip)'
SERVICE_URL = 'http://www.librarything.com/api/thingISBN/{isbn}'


def _get_text(topnode):
    """Get the text values in the child nodes."""
    text = ""
    for node in topnode.childNodes:
        if node.nodeType == node.TEXT_NODE:
            text = text + node.data
    return text


def parser_thinged(xml):
    """Parse the response from the ThingISBN service."""
    dom = parseString(xml)
    nodes = dom.getElementsByTagName("idlist")[0].getElementsByTagName("isbn")
    return [EAN13(_get_text(isbn)) for isbn in nodes]


def query(isbn):
    """Query the worldcat.org service for related ISBNs."""
    data = wquery(SERVICE_URL.format(isbn=isbn),
                  user_agent=UA,
                  parser=parser_thinged)
    if not data:           # pragma: no cover
        LOGGER.debug('No data from ThingISBN for isbn %s', isbn)
    return data



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

from nose.tools import assert_equals
from .._core import (_check_digit10, _check_digit13, _check_structure10,
    _check_structure13, is_isbn10, is_isbn13, to_isbn10, to_isbn13,
    canonical, clean, notisbn, get_isbnlike, get_canonical_isbn, EAN13)
from .data4tests import ISBNs


# nose tests

def test__check_digit10():
    """Test check digit algo for ISBN-10."""
    assert_equals(_check_digit10('082649752'), '7')
    assert_equals(_check_digit10('585270001'), '0')
    assert_equals(_check_digit10('08264975X'), None)
    assert_equals(_check_digit10('08264975'), None)


def test__check_digit13():
    """Test check digit algo for ISBN-13."""
    assert_equals(_check_digit13('978082649752'), '9')
    assert_equals(_check_digit13('97808264975'), None)
    assert_equals(_check_digit13('97808264975X'), None)


def test__check_structure10():
    """Test structure detection for ISBN-10."""
    assert_equals(_check_structure10('0826497527'), True)
    assert_equals(_check_structure10('0826497X27'), True) # isbnlike!
    assert_equals(_check_structure10('0826497XI7'), False)


def test__check_structure13():
    """Test structure detection for ISBN-13."""
    assert_equals(_check_structure13('9780826497529'), True)
    assert_equals(_check_structure13('978082649752X'), False)


def test_is_isbn10():
    """Test detection and validation for ISBN-10."""
    assert_equals(is_isbn10('0826497527'), True)
    assert_equals(is_isbn10('isbn 0-8264-9752-7'), True)
    assert_equals(is_isbn10('0826497520'), False)
    assert_equals(is_isbn10('954430603X'), True)


def test_is_isbn13():
    """Test detection and validation for ISBN-13."""
    assert_equals(is_isbn13('9780826497529'), True)
    assert_equals(is_isbn13('9791090636071'), True)
    assert_equals(is_isbn13('isbn 979-10-90636-07-1'), True)
    assert_equals(is_isbn13('9780826497520'), False)
    assert_equals(is_isbn13('9700000000000'), False)
    assert_equals(is_isbn13('9000000000000'), False)
    assert_equals(is_isbn13('9710000000000'), False)


def test_to_isbn10():
    """Test transformation of ISBN to ISBN-10."""
    assert_equals(to_isbn10('9780826497529'), '0826497527')
    assert_equals(to_isbn10('0826497527'), '0826497527')
    assert_equals(to_isbn10('9780826497520'), None)     # ISBN13 not valid
    assert_equals(to_isbn10('9790826497529'), None)
    assert_equals(to_isbn10('97808264975X3'), None)
    assert_equals(to_isbn10('978-826497'), None)               # (bug #14)
    assert_equals(to_isbn10('isbn 0-8264-9752-7'), '0826497527')
    assert_equals(to_isbn10('isbn 979-10-90636-07-1'), None)
    assert_equals(to_isbn10('isbn 978-0-8264-9752-9'), '0826497527')
    assert_equals(to_isbn10('asdadv isbn 978-0-8264-9752-9'), '0826497527')


def test_to_isbn13():
    """Test transformation of ISBN to ISBN-13."""
    assert_equals(to_isbn13('0826497527'), '9780826497529')
    assert_equals(to_isbn13('9780826497529'), '9780826497529')
    assert_equals(to_isbn13('0826497520'), None)        # ISBN10 not valid
    assert_equals(to_isbn13('08X6497527'), None)
    assert_equals(to_isbn13('91-43-01019-9'), '9789143010190') # (bug #14)
    assert_equals(to_isbn13('isbn 91-43-01019-9'), '9789143010190')
    assert_equals(to_isbn13('asd isbn 979-10-90636-07-1 blabla'), '9791090636071')


def test_clean():
    """Test the cleanning of ISBN-like strings."""
    assert_equals(clean(' 978.0826.497529'), '9780826497529')
    assert_equals(clean('ISBN: 9791090636071'), 'ISBN 9791090636071')
    assert_equals(clean('978,0826497520'), '9780826497520')


def test_notisbn():
    """Test the impossibility of extracting valid ISBN from ISBN-like strings."""
    assert_equals(notisbn('0826497527'), False)
    assert_equals(notisbn('0826497520'), True)
    assert_equals(notisbn('9780826497529', level='strict'), False)
    assert_equals(notisbn('9426497529', level='strict'), True)
    assert_equals(notisbn('978082649752', level='strict'), True)
    assert_equals(notisbn('978082649752', level='loose'), True)
    assert_equals(notisbn('9780826400001', level='loose'), False)
    assert_equals(notisbn('9780826400001', level='strict'), True)
    assert_equals(notisbn('9780826400001', level='badlevel'), None)
    assert_equals(notisbn('978 9426497529'), True)
    assert_equals(notisbn('9789426497529'), True)
    assert_equals(notisbn('979 10 9063607 1'), False)
    assert_equals(notisbn('9780826497520'), True)


def test_get_isbnlike():
    """Test the extraction of ISBN-like strings."""
    assert_equals(len(get_isbnlike(ISBNs)), 79)
    assert_equals(len(get_isbnlike(ISBNs, 'normal')), 79)
    assert_equals(len(get_isbnlike(ISBNs, 'strict')), 69)
    assert_equals(len(get_isbnlike(ISBNs, 'loose')), 81)
    assert_equals(get_isbnlike(ISBNs, 'e'), None)


def test_get_canonical_isbn():
    """Test the extraction of canonical ISBN from ISBN-like string."""
    assert_equals(get_canonical_isbn('0826497527', output='bouth'),
                  '0826497527')
    assert_equals(get_canonical_isbn('0826497527'), '0826497527')
    assert_equals(get_canonical_isbn('0826497527', output='isbn10'),
                  '0826497527')
    assert_equals(get_canonical_isbn('0826497527', output='isbn13'),
                  '9780826497529')
    assert_equals(get_canonical_isbn('ISBN 0826497527', output='isbn13'),
                  '9780826497529')
    assert_equals(get_canonical_isbn('ISBN 0826497527', output='NOOPTION'),
                  None)
    assert_equals(get_canonical_isbn('0826497520'), None)
    assert_equals(get_canonical_isbn('9780826497529'), '9780826497529')
    assert_equals(get_canonical_isbn('9780826497520'), None)
    assert_equals(get_canonical_isbn('OSX 9780826497529.pdf'), '9780826497529')


def test_canonical():
    """Test the extraction of 'only numbers and X' from ISBN-like string."""
    assert_equals(canonical('ISBN 9789720404427'), '9789720404427')
    assert_equals(canonical('ISBN-9780826497529'), '9780826497529')
    assert_equals(canonical('ISBN9780826497529'), '9780826497529')
    assert_equals(canonical('isbn9780826497529'), '9780826497529')
    assert_equals(canonical('isbn 0826497527'), '0826497527')
    assert_equals(canonical('954430603x'), '954430603X')


def test_EAN13():
    """Test the extraction and validation of EAN13 from ISBN-like string."""
    assert_equals(EAN13('ISBN 9789720404427'), None)
    assert_equals(EAN13('ISBN 9789720404428'), '9789720404428')
    assert_equals(EAN13('ISBN-9780826497529'), '9780826497529')
    assert_equals(EAN13('ISBN9780826497529'), '9780826497529')
    assert_equals(EAN13('isbn9780826497529'), '9780826497529')
    assert_equals(EAN13('isbn 0826497527'), '9780826497529')
    assert_equals(EAN13('9700000000000'), None)
    assert_equals(EAN13('9000000000000'), None)
    assert_equals(EAN13('9710000000000'), None)



# -*- coding: utf-8 -*-

# isbnlib - tools for extracting, cleaning and transforming ISBNs
# Copyright (C) 2014  Alexandre Lima Conde

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""isbnlib main file.

Tools for extracting, cleaning, transforming and validating ISBN ids.
"""

import logging
import re

LOGGER = logging.getLogger(__name__)


RE_ISBN10 = re.compile(r'ISBN\x20(?=.{13}$)\d{1,5}([- ])\d{1,7}'
                       r'\1\d{1,6}\1(\d|X)$|[- 0-9X]{10,16}')
RE_ISBN13 = re.compile(r'97[89]{1}(?:-?\d){10,16}|97[89]{1}[- 0-9]{10,16}')
RE_STRICT = re.compile(r'^(?:ISBN(?:-1[03])?:? )?(?=[0-9X]{10}$|'
                       r'(?=(?:[0-9]+[- ]){3})'
                       r'[- 0-9X]{13}$|97[89][0-9]{10}$|'
                       r'(?=(?:[0-9]+[- ]){4})'
                       r'[- 0-9]{17}$)(?:97[89][- ]?)?[0-9]{1,5}'
                       r'[- ]?[0-9]+[- ]?[0-9]+[- ]?[0-9X]$',
                       re.I | re.M | re.S)
RE_NORMAL = re.compile(r'97[89]{1}(?:-?\d){10}|\d{9}[0-9X]{1}|'
                       r'[-0-9X]{10,16}',
                       re.I | re.M | re.S)
RE_LOOSE = re.compile(r'[- 0-9X]{10,19}', re.I | re.M | re.S)
ISBN13_PREFIX = '978'
LEGAL = '0123456789xXisbnISBN- '


def _check_digit10(firstninedigits):
    """Check sum ISBN-10."""
    # minimum checks
    if len(firstninedigits) != 9:
        return None
    try:
        int(firstninedigits)
    except:  # pragma: no cover
        return None
    # checksum
    val = sum((i + 2) * int(x) for
              i, x in enumerate(reversed(firstninedigits)))
    remainder = int(val % 11)
    if remainder == 0:
        tenthdigit = 0
    else:
        tenthdigit = 11 - remainder
    if tenthdigit == 10:
        tenthdigit = 'X'
    return str(tenthdigit)


def _check_digit13(firsttwelvedigits):
    """Check sum ISBN-13."""
    # minimum checks
    if len(firsttwelvedigits) != 12:
        return None
    try:
        int(firsttwelvedigits)
    except:  # pragma: no cover
        return None
    # checksum
    val = sum((i % 2 * 2 + 1) * int(x) for
              i, x in enumerate(firsttwelvedigits))
    thirteenthdigit = 10 - int(val % 10)
    if thirteenthdigit == 10:
        thirteenthdigit = '0'
    return str(thirteenthdigit)


def _check_structure10(isbn10like):
    """Check structure of an ISBN-10."""
    return True if re.match(RE_ISBN10, isbn10like) else False


def _check_structure13(isbn13like):
    """Check structure of an ISBN-13."""
    return True if re.match(RE_ISBN13, isbn13like) else False


def is_isbn10(isbn10):
    """Validate as ISBN-10."""
    isbn10 = canonical(isbn10)
    if len(isbn10) != 10:
        return False          # pragma: no cover
    else:
        return False if _check_digit10(isbn10[:-1]) != isbn10[-1] else True


def is_isbn13(isbn13):
    """Validate as ISBN-13."""
    isbn13 = canonical(isbn13)
    if len(isbn13) != 13:
        return False          # pragma: no cover
    else:
        if isbn13[0:3] not in ('978', '979'):
            return False
        return False if _check_digit13(isbn13[:-1]) != isbn13[-1] else True


def to_isbn10(isbn13):
    """Transform isbn-13 to isbn-10."""
    isbn13 = canonical(isbn13)
    # Check prefix
    if isbn13[:3] != ISBN13_PREFIX:
        return isbn13 if len(isbn13) == 10 and is_isbn10(isbn13) else None
    if not is_isbn13(isbn13):
        return None
    isbn10 = isbn13[3:]
    check = _check_digit10(isbn10[:-1])
    # Change check digit
    return isbn10[:-1] + check if check else None


def to_isbn13(isbn10):
    """Transform isbn-10 to isbn-13."""
    isbn10 = canonical(isbn10)
    if len(isbn10) == 13 and is_isbn13(isbn10):
        return isbn10
    if not is_isbn10(isbn10):
        return None
    isbn13 = ISBN13_PREFIX + isbn10[:-1]
    check = _check_digit13(isbn13)
    return isbn13 + check if check else None


def canonical(isbnlike):
    """Keep only numbers and X."""
    numb = [c for c in isbnlike if c in '0123456789Xx']
    if numb and numb[-1] == 'x':
        numb[-1] = 'X'
    return ''.join(numb)


def clean(isbnlike):
    """Clean ISBN (only legal characters)."""
    cisbn = [c for c in isbnlike if c in LEGAL]
    buf = re.sub(r'\s*-\s*', '-', ''.join(cisbn))
    return re.sub(r'\s+', ' ', buf).strip()


def notisbn(isbnlike, level='strict'):
    """Check with the goal to invalidate isbn-like.

    level:
    strict for certain they are not ISBNs (default)
    loose  only filters obvious NO ISBNs

    """
    if level not in ('strict', 'loose'):  # pragma: no cover
        LOGGER.error('level as no option %s', level)
        return
    isbnlike = canonical(isbnlike)
    if len(isbnlike) not in (10, 13):
        return True
    if level != 'strict':
        return False
    if len(isbnlike) == 10:
        return not is_isbn10(isbnlike)
    else:
        return not is_isbn13(isbnlike)


def get_isbnlike(text, level='normal'):
    """Extract all substrings that seem like ISBNs.

    level:
    strict almost as certain they are ISBNs
    normal (default)
    loose  catch many as possible

    """
    if level == 'normal':  # pragma: no cover
        isbnlike = RE_NORMAL
    elif level == 'strict':
        isbnlike = RE_STRICT
    elif level == 'loose':
        isbnlike = RE_LOOSE
    else:
        LOGGER.error('level as no option %s', level)
        return
    return isbnlike.findall(text)


def get_canonical_isbn(isbnlike, output='bouth'):
    """Extract ISBNs and transform them to the canonical form.

    output:
    isbn10
    isbn13
    bouth  (default)

    """
    if output not in ('bouth', 'isbn10', 'isbn13'):  # pragma: no cover
        LOGGER.error('output as no option %s', output)
        return

    regex = RE_NORMAL

    match = regex.search(isbnlike)
    if match:
        # Get only canonical characters and split them into a list
        cisbn = canonical(match.group())
        chars = list(cisbn)
        # Remove the last digit from `chars` and assign it to `last`
        last = chars.pop()
        buf = ''.join(chars)

        if len(chars) == 9:
            # Compute the ISBN-10 checksum digit
            check = _check_digit10(buf)
        else:
            # Compute the ISBN-13 checksum digit
            check = _check_digit13(buf)

        # If checksum OK return a `canonical` ISBN
        if str(check) == last:
            if output == 'bouth':
                return cisbn
            if output == 'isbn10':
                return cisbn if len(cisbn) == 10 else to_isbn10(cisbn)
            return to_isbn13(cisbn) if len(cisbn) == 10 else cisbn


def ean13(isbnlike):
    """Transform an `isbnlike` string in an EAN number (canonical ISBN-13)."""
    ib = canonical(isbnlike)
    if len(ib) == 13:
        return ib if is_isbn13(ib) else None
    elif len(ib) == 10:
        return to_isbn13(ib) if is_isbn10(ib) else None

# Alias
EAN13 = ean13



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""
nose tests
"""

from .. import _goom as goom
from nose.tools import assert_equals


def test_goom():
    """Test the Google's Multiple Books service."""
    assert_equals(len(repr(goom.query('the old man and the sea'))) > 1000, True)



# -*- coding: utf-8 -*-

""" Exceptions for isbnlib.dev

The classes in isbnlib.dev should use the exceptions below.
"""


class ISBNLibDevException(Exception):
    """ Base class for isbnlib.dev exceptions

    This exception should not be raised directly,
    only subclasses of this exception should be used!
    """

    def __init__(self, msg=None):
        if msg:
            self.message = '%s (%s)' % (self.message, msg)

    def __str__(self):
        return getattr(self, 'message', '')       # pragma: no cover


class ISBNLibHTTPError(ISBNLibDevException):
    """ Exception raised for HTTP related errors
    """
    message = "an HTTP error has ocurred"


class ISBNLibURLError(ISBNLibDevException):
    """ Exception raised for URL related errors
    """
    message = "an URL error has ocurred"


class DataNotFoundAtServiceError(ISBNLibDevException):
    """ Exception raised when there is no target data from the service
    """
    message = "the target data was not found at this service"


class ServiceIsDownError(ISBNLibDevException):
    """ Exception raised when the service is not available
    """
    message = "the service is down (try later)"


class DataWrongShapeError(ISBNLibDevException):
    """ Exception raised when the data hasn't the expected format
    """
    message = "the data hasn't the expected format"


class NoDataForSelectorError(ISBNLibDevException):
    """ Exception raised when there is no data for the selector
    """
    message = "no data for this selector"


class NotValidMetadataError(ISBNLibDevException):
    """ Exception raised when the metadata hasn't the expected format
    """
    message = "the metadata hasn't the expected format"


class ISBNNotConsistentError(ISBNLibDevException):
    """ Exception raised when the isbn request != isbn response
    """
    message = "isbn request != isbn response"


class RecordMappingError(ISBNLibDevException):
    """ Exception raised when the mapping records -> canonical doesn't work
    """
    message = "the mapping `canonical <- records` doesn't work"


class NoAPIKeyError(ISBNLibDevException):
    """ Exception raised when the API Key for a service is not found
    """
    message = "this service needs an API key"


class FileNotFoundError(ISBNLibDevException):
    """ Exception raised when a given file doesn't exist
    """
    message = "the file wasn't found"



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file

"""
nose tests
"""

from .. import _gwords as words
from nose.tools import assert_equals


def test_words():
    """Test 'isbn_from_words' function."""
    assert_equals(len(words.goos('the old man and the sea')), 13)
    # assert_equals(words.goos('-ISBN -isbn') in ('9781364200329', None), True)



# -*- coding: utf-8 -*-
"""Hyphenate an ISBN."""

import logging

from ._core import EAN13, canonical, to_isbn13
from ._data.data4mask import ranges
from ._exceptions import NotValidISBNError

LOGGER = logging.getLogger(__name__)


def msk(isbn, separator='-'):
    """Transform a canonical ISBN to a `masked` one.

    `Mask` the ISBN, separating by identifier
    ISBN-10 identifiers: country-publisher-title-check

    Used the iterative version of the `sliding-window` algorithm.
    Not pretty, but fast! Lines 42-52 implement the search loop.
    O(n) for n (number of keys),
    if data structure like 'ranges' in data4mask.py
    """
    ib = canonical(isbn)
    ean = EAN13(ib)
    if len(ib) not in (10, 13) or ean is None:
        LOGGER.critical('%s is not a valid ISBN', isbn)
        raise NotValidISBNError(isbn)

    isbn10 = False
    if len(ib) == 10:
        check10 = ib[-1]
        ib = to_isbn13(ib)
        isbn10 = True

    idx = None
    check = ib[-1:]                     # <-- hack!
    cur = 3
    igi = cur
    buf = ib[igi:cur + 1]
    group = ib[0:cur] + '-' + buf

    for _ in range(6):                             # pragma: no cover
        if group in ranges:
            sevens = ib[cur + 1:cur + 8].ljust(7, '0')
            for l in ranges[group]:
                if l[0] <= int(sevens) <= l[1]:
                    idx = l[2]
                    break
            break
        cur += 1
        buf = ib[igi:cur + 1]
        group = group + buf[-1:]        # <-- hack!

    if idx:
        if isbn10:
            group = group[4:]
            check = check10
        return separator.join([group, ib[cur + 1:cur + idx + 1],
                              ib[cur + idx + 1:-1], check])
    LOGGER.warning('identifier not found! '
                   'Please, update the program.')  # pragma: no cover
    return                                         # pragma: no cover



# -*- coding: utf-8 -*-

"""Read and write cover cache.


    NOTE
    1. shelve has different incompatible formats in py2 and py3
    2. if some methods detect that the cache is not consistent
       they delete the cache and create a new one.
    3. After purge the cache keeps the records with more hits
       and the newests.
    4. The design is for safety not for performance! Increasing
       MAXLEN can have an high detrimental impact on performance.

    Examples:
    cc = CoversCache('.covers')
    cc['9781408835029'] = (
        "http://books.google.com/books/content?id=uUcvgfTYTRnjh"\
        "&printsec=frontcover"\
        "&img=1&zoom=2&edge=curl&source=gbs_api",
        ".covers/slot01/9781408835029.jpg"
        )
    cc['9781408835029']
    cc.hits('9781408835029')
    cc.keys()
    cc.files()
    cc.sync()
    cc.make()
    cc.delete()
    del cc['9781408835029']

"""

import os
import shutil
import sys
from random import randint

from ._shelvecache import ShelveCache

WINDOWS = os.name == 'nt'
PY3 = sys.version > '3'


class CoversCache(object):

    """Covers cache."""

    CACHEFOLDER = '.covers'
    INDEXFN = '.index'
    WINSHADOWIDX = INDEXFN + '.dat'
    MAXLEN = 3000
    NSLOTS = 10

    def __init__(self, cachepath=CACHEFOLDER):
        """Initialize attributes."""
        self.cachepath = cachepath
        self._indexpath = os.path.join(cachepath, self.INDEXFN)
        if WINDOWS and PY3:
            if not os.path.isfile(os.path.join(cachepath, self.WINSHADOWIDX)):
                self.make()
        elif not os.path.isfile(self._indexpath):
            self.make()
        self._index = ShelveCache(self._indexpath)
        self._index.MAXLEN = self.MAXLEN
        if len(self._index) > self.MAXLEN:
            self.purge()

    def __getitem__(self, key):
        """Read cache."""
        try:
            url, pth = self._index[key]
            pth = os.path.join(self.cachepath, pth)
            return (url, pth)
        except:
            return None

    def __setitem__(self, key, value):
        """Write to cache."""
        url, pth = value
        try:
            if pth and os.path.isfile(pth) and url:
                tocache = os.path.join(self._get_slot(), os.path.basename(pth))
                target = os.path.join(self.cachepath, tocache)
                shutil.copyfile(pth, target)
                if os.path.isfile(target):
                    self._index[key] = (url, tocache)
                    return True
                return False
            else:
                raise
        except:
            return False

    def __delitem__(self, key):
        """Delete record with key."""
        try:
            del self._index[key]
            return True
        except:
            return False

    def __len__(self):
        """Return the number of keys in cache."""
        return len(self._index.keys()) if self._index.keys() else 0

    def keys(self):
        """Return the number of keys in cache."""
        return self._index.keys()

    def hits(self, key):
        """Return the number of hits for key."""
        return self._index.hits(key)

    def _create_slots(self):
        for slot in range(self.NSLOTS):
            name = "slot%02d" % slot
            pth = os.path.join(self.cachepath, name)
            if not os.path.exists(pth):
                os.mkdir(pth)

    def make(self):
        """Init the cache."""
        # 1. Delete if available
        if os.path.isdir(self.cachepath):
            self.delete()
        # 2. Make folder
        os.mkdir(self.cachepath)
        # 3. Create Index
        self._index = ShelveCache(self._indexpath)
        # 4. Create slots
        self._create_slots()

    def _get_slot(self):
        return "slot%02d" % randint(0, self.NSLOTS - 1)

    def files(self):
        pths = []
        for root, _, fls in os.walk(self.cachepath):
            for fn in fls:
                pths.append(os.path.join(root, fn))
        return pths

    def sync(self):
        """Sync index entries with disk files."""
        # clear index entries not on disk
        checked = [self._indexpath]
        for key in self._index.keys():
            url, pth = self._index[key]
            pth = os.path.join(self.cachepath, pth)
            if not os.path.isfile(pth):
                self._index[key] = (url, None)
            checked.append(pth)
        # delete files not on index
        diff = tuple(set(self.files()) - set(checked))
        for fp in diff:
            if self.INDEXFN not in fp:
                os.remove(fp)

    def purge(self):
        try:
            self._index.purge()
            self.sync()
            return True
        except:
            return False

    def delete(self):
        self._index = None
        return shutil.rmtree(self.cachepath)



# -*- coding: utf-8 -*-
"""Query the worldcat.org xID service for metadata."""

import logging

from .dev import stdmeta
from .dev._bouth23 import u
from .dev._exceptions import (DataWrongShapeError, ISBNNotConsistentError,
                              NoDataForSelectorError, RecordMappingError)
from .dev.webquery import query as wquery

UA = 'isbnlib (gzip)'
SERVICE_URL = 'http://xisbn.worldcat.org/webservices/xid/isbn/{isbn}?'\
    'method=getMetadata&format=json&fl=title,author,year,publisher,lang'
LOGGER = logging.getLogger(__name__)


def _mapper(isbn, records):
    """Mapp: canonical <- records."""
    # canonical: ISBN-13, Title, Authors, Publisher, Year, Language
    try:
        canonical = {}
        canonical['ISBN-13'] = u(isbn)
        canonical['Title'] = records.get('title', u('')).replace(' :', ':')
        buf = records.get('author', u(''))
        canonical['Authors'] = [x.strip('. ') for x in buf.split(';')]
        canonical['Publisher'] = records.get('publisher', u(''))
        canonical['Year'] = records.get('year', u(''))
        canonical['Language'] = records.get('lang', u(''))
    except:  # pragma: no cover
        LOGGER.debug("RecordMappingError for %s with data %s", isbn, records)
        raise RecordMappingError(isbn)
    # call stdmeta for extra cleanning and validation
    return stdmeta(canonical)


def _records(isbn, data):
    """Classify (canonically) the parsed data."""
    # check status
    try:
        status = data['stat']
        if status != 'ok':
            raise
    except:
        LOGGER.debug('DataWrongShapeError for %s with status %s',
                     isbn, status)
        raise DataWrongShapeError("status: '%s' for isbn %s" % (status, isbn))
    # put the selected data in records
    try:
        recs = data['list'][0]
    except:  # pragma: no cover
        LOGGER.debug('NoDataForSelectorError for %s', isbn)
        raise NoDataForSelectorError(isbn)
    # consistency check (isbn request = isbn response)
    if recs:
        ids = recs.get('isbn', '')
        if isbn not in repr(ids):   # pragma: no cover
            LOGGER.debug('ISBNNotConsistentError for %s (%s)', isbn, repr(ids))
            raise ISBNNotConsistentError("%s not in %s" % (isbn, repr(ids)))
    # map canonical <- records
    return _mapper(isbn, recs)


def query(isbn):
    """Query the worldcat.org service for metadata."""
    data = wquery(SERVICE_URL.format(isbn=isbn), user_agent=UA)
    return _records(isbn, data)



# -*- coding: utf-8 -*-
"""Query the Google Books (JSON API v1) for metadata."""

import logging

from .dev import stdmeta
from .dev._bouth23 import u
from .dev._exceptions import (NoDataForSelectorError, RecordMappingError)
from .dev.webquery import query as wquery

UA = 'isbnlib (gzip)'
SERVICE_URL = 'https://www.googleapis.com/books/v1/volumes?q={words}'\
    '&fields=items/volumeInfo(title,authors,publisher,publishedDate,'\
    'language,industryIdentifiers)&maxResults=10'
LOGGER = logging.getLogger(__name__)


def _mapper(record):
    """Map canonical <- record."""
    # canonical:
    # -> ISBN-13, Title, Authors, Publisher, Year, Language
    try:
        # mapping: canonical <- records
        if 'industryIdentifiers' not in record:  # pragma: no cover
            return
        canonical = {}
        isbn = None
        for ident in record['industryIdentifiers']:
            if ident['type'] == 'ISBN_13':
                isbn = ident['identifier']
                break
        if not isbn:  # pragma: no cover
            return
        canonical['ISBN-13'] = isbn
        canonical['Title'] = record.get('title', u('')).replace(' :', ':')
        canonical['Authors'] = record.get('authors', [])
        canonical['Publisher'] = record.get('publisher', u(''))
        if 'publishedDate' in record \
           and len(record['publishedDate']) >= 4:
            canonical['Year'] = record['publishedDate'][0:4]
        else:         # pragma: no cover
            canonical['Year'] = u('')
        canonical['Language'] = record.get('language', u(''))
    except:           # pragma: no cover
        raise RecordMappingError(isbn)
    # call stdmeta for extra cleanning and validation
    return stdmeta(canonical)


def _records(words, data):
    """Classify (canonically) the parsed data."""
    # put the selected data in records
    try:
        recs = [d['volumeInfo'] for d in data['items']]
    except:  # pragma: no cover
        LOGGER.debug('NoDataForSelectorError for (%s)', words)
        raise NoDataForSelectorError(words)
    # map canonical <- records
    return [_mapper(r) for r in recs if _mapper(r)]


def query(words):
    """Query the Google Books (JSON API v1) for metadata."""
    data = wquery(SERVICE_URL.format(words=words.replace(' ', '+')), UA)
    return _records(words, data)



# -*- coding: utf-8 -*-
from . import bouth23, helpers, vias
from ._data import Metadata, stdmeta
from ._exceptions import (DataNotFoundAtServiceError, DataWrongShapeError,
                          ISBNLibDevException, ISBNLibHTTPError,
                          ISBNLibURLError, NoAPIKeyError,
                          NoDataForSelectorError, NotValidMetadataError,
                          RecordMappingError, ServiceIsDownError)
from .webquery import WEBQuery
from .webservice import WEBService

# alias
ISBNToolsDevException = ISBNLibDevException
ISBNToolsHTTPError = ISBNLibHTTPError
ISBNToolsURLError = ISBNLibURLError

__all__ = ('ISBNToolsDevException',
           'ISBNLibDevException',
           'ISBNToolsHTTPError', 'ISBNToolsURLError',
           'ISBNLibHTTPError', 'ISBNLibURLError',
           'DataNotFoundAtServiceError',
           'ServiceIsDownError', 'DataWrongShapeError',
           'NotValidMetadataError',
           'NoDataForSelectorError',
           'RecordMappingError', 'NoAPIKeyError',
           'Metadata', 'stdmeta',
           'WEBService', 'WEBQuery', 'vias', 'bouth23', 'helpers'
           )



# -*- coding: utf-8 -*-

"""Read and write shelve cache.

NOTES
1. shelve has different incompatible formats in py2 and py3.
2. If some methods detect that the cache is not consistent
   they delete the cache and create a new one.
3. After purge, the cache keeps the records with more hits
   and the newests.
4. By opening and closing in each operation, the cache performs badly
   for many records (because it doesn't use the 'in memory' part of cache).
   So don't increase MAXLEN too much.
5. The cache is optimized for low hit frequency (using a simple dict lookup
   not a Bloom filter!).

"""

# TODO rewrite using context managers!

import datetime
import shelve
from time import time as timestamp


class ShelveCache(object):

    """Read and write shelve cache."""

    MAXLEN = 2000
    CUTOFF = 0.5

    def __init__(self, filepath):
        """Initialize attributes."""
        self._sh = shelve
        self.filepath = filepath
        try:
            s = self._sh.open(self.filepath)
            try:
                self._keys = list(s.keys())
                if len(self._keys) > self.MAXLEN:
                    self.purge()
            except:
                pass
        except:
            s = self._sh.open(self.filepath, 'n')
            self._keys = []
        s.close()

    def __getitem__(self, key):
        """Read cache."""
        if key not in self._keys:
            return None
        try:
            s = self._sh.open(self.filepath, writeback=True)
            if s[key]:
                s[key]['hits'] += 1
                return s[key]['value']
            else:
                return None
        except KeyError:
            return None
        except ValueError:
            self.new()
            return None
        finally:
            s.close()

    def __setitem__(self, key, value):
        """Write to cache."""
        try:
            s = self._sh.open(self.filepath)
            s[key] = {'value': value, 'hits': 0, 'timestamp': timestamp()}
            self._keys.append(key)
            status = True
        except:
            status = False
        finally:
            s.close()
        return status

    def __delitem__(self, key):
        """Delete record with key."""
        if key not in self._keys:
            return
        try:
            s = self._sh.open(self.filepath)
            del s[key]
            self._keys.remove(key)
        except KeyError:
            return
        except ValueError:
            self.new()
            return
        finally:
            s.close()

    def __len__(self):
        """Return the number of keys in cache."""
        return len(self.keys()) if self.keys() else 0

    def keys(self):
        """Return list of keys in Cache."""
        if self._keys:
            return self._keys
        try:
            s = self._sh.open(self.filepath)
            self._keys = list(s.keys())
            return self._keys
        finally:
            s.close()

    def ts(self, key):
        """Return the timestamp of the record with key."""
        if key not in self._keys:
            return None
        try:
            s = self._sh.open(self.filepath)
            ts = s[key]['timestamp'] if s[key] else None
            if not ts:
                return
            fmt = '%Y-%m-%d %H:%M:%S'
            return datetime.datetime.fromtimestamp(ts).strftime(fmt)
        except KeyError:
            return
        except ValueError:
            self.new()
            return
        finally:
            s.close()

    def hits(self, key):
        """Return the number of hits for the record with key."""
        if key not in self._keys:
            return None
        try:
            s = self._sh.open(self.filepath)
            hts = s[key]['hits'] if s[key] else None
            return hts
        except KeyError:
            return
        except ValueError:
            self.new()
            return
        finally:
            s.close()

    def new(self):
        """Make new cache."""
        s = self._sh.open(self.filepath, 'n')
        s.close()
        self._keys = []

    def purge(self):
        """Purge the cache."""
        try:
            if len(self.keys()) < self.MAXLEN:
                return
            s = self._sh.open(self.filepath)
            data = [(k, s[k]['timestamp'], s[k]['hits']) for k in s.keys()]
            data.sort(key=lambda tup: (-tup[2], -tup[1]))
            keep = int(self.CUTOFF * self.MAXLEN)
            garbk = [k[0] for k in data[keep:]]
            for k in garbk:
                del s[k]
            self._keys = s.keys()
        finally:
            s.close()



# -*- coding: utf-8 -*-

# isbnlib - tools for extracting, cleaning and transforming ISBNs
# Copyright (C) 2014  Alexandre Lima Conde

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from setuptools import setup
from isbnlib import __version__


setup(
    name='isbnlib',
    version=__version__,
    author='xlcnd',
    author_email='xlcnd@outlook.com',
    url='https://github.com/xlcnd/isbnlib',
    download_url='https://github.com/xlcnd/isbnlib/archive/v3.5.6.zip',
    packages=['isbnlib',
              'isbnlib/dev',
              'isbnlib/_data',
              'isbnlib/test',
              ],
    license='LGPL v3',
    description='Extract, clean, transform, hyphenate and metadata for ISBNs (International Standard Book Number).',
    long_description=open('README.rst').read(),
    keywords='ISBN, validate, transform, hyphenate, metadata, World Catalogue, Google Books, Open Library, isbndb.com, BibTeX, EndNote, RefWorks, MSWord, opf, BibJSON, ISBN-A, doi, EAN13',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Operating System :: OS Independent',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Text Processing :: General',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    tests_require=['nose'],
    test_suite='nose.collector',
)



# -*- coding: utf-8 -*-
# flake8:noqa
# pylint: skip-file
"""Expose usefull features."""

from .._imcache import IMCache
from ._coverscache import CoversCache
from ._files import File, cwdfiles
from ._fmt import fmtbib, fmts
from ._helpers import unicode_to_utf8tex as to_utf8tex
from ._helpers import (cutoff_tokens, fake_isbn, in_virtual, last_first,
                       normalize_space, parse_placeholders)
from ._shelvecache import ShelveCache


__all__ = ['CoversCache', 'File', 'IMCache', 'ShelveCache', 'cutoff_tokens', 'cwdfiles',
           'fmtbib', 'fmts', 'in_virtual', 'last_first', 'normalize_space',
           'parse_placeholders', 'to_utf8tex', 'fake_isbn'
           ]



# -*- coding: utf-8 -*-
"""Query the isbndb.org service for metadata."""

import logging
import re

from .config import apikeys
from .dev import stdmeta
from .dev._bouth23 import u
from .dev._exceptions import (DataWrongShapeError, ISBNNotConsistentError,
                              NoAPIKeyError, NoDataForSelectorError,
                              RecordMappingError)
from .dev.webquery import query as wquery

UA = 'isbnlib (gzip)'
SERVICE_URL = 'http://isbndb.com/api/v2/json/{apikey}/book/{isbn}'
PATT_YEAR = re.compile(r'\d{4}')
LOGGER = logging.getLogger(__name__)


def _mapper(isbn, records):
    """Map canonical <- records."""
    # canonical:
    # -> ISBN-13, Title, Authors, Publisher, Year, Language
    try:
        # mapping: canonical <- records
        canonical = {}
        canonical['ISBN-13'] = u(isbn)
        # assert isbn == records['isbn13'], "isbn was mungled!"
        canonical['Title'] = records.get('title', u(''))
        authors = [a['name'] for a in records['author_data']]
        canonical['Authors'] = authors
        canonical['Publisher'] = records.get('publisher_name', u(''))
        canonical['Year'] = u('')
        if 'edition_info' in records:
            match = re.search(PATT_YEAR, records['edition_info'])
            if match:
                canonical['Year'] = str(match.group(0))
        canonical['Language'] = records.get('language', u(''))
    except:
        raise RecordMappingError(isbn)
    # call stdmeta for extra cleanning and validation
    return stdmeta(canonical)


def _records(isbn, data):
    """Classify (canonically) the parsed data."""
    # check status
    try:
        err = data.get('error', '')
        if err:
            raise
    except:
        LOGGER.debug('DataWrongShapeError for %s with status %s',
                     isbn, err)
        raise DataWrongShapeError("error: '%s' for isbn %s" % (err, isbn))
    # put the selected data in records
    try:
        recs = data['data'][0]
    except:  # pragma: no cover
        LOGGER.debug('NoDataForSelectorError for %s', isbn)
        raise NoDataForSelectorError(isbn)
    # consistency check (isbn request = isbn response)
    if recs:
        ids = recs.get('isbn13', '')
        if isbn not in repr(ids):   # pragma: no cover
            LOGGER.debug('ISBNNotConsistentError for %s (%s)', isbn, repr(ids))
            raise ISBNNotConsistentError("%s not in %s" % (isbn, repr(ids)))
    # map canonical <- records
    return _mapper(isbn, recs)


def query(isbn):
    """Query the isbndb.org service for metadata."""
    if not apikeys.get('isbndb'):
        raise NoAPIKeyError
    data = wquery(SERVICE_URL.format(apikey=apikeys['isbndb'], isbn=isbn),
                  user_agent=UA)
    return _records(isbn, data)



# -*- coding: utf-8 -*-
# flake8: noqa
# pylint: skip-file


""" nose tests
"""

from .._metadata import query
from nose.tools import assert_equals, assert_raises


def test_query():
    """Test 'openl' metadata service."""
    # test query from metadata
    assert_equals(len(repr(query('9780195132861', 'openl'))) > 140, True)
    assert_equals(len(repr(query('9780156001311', 'openl'))) > 140, True)



# -*- coding: utf-8 -*-
"""Get the cover of the book."""

import logging

try:                     # pragma: no cover
    # PY3
    from urllib.error import HTTPError, URLError
    from urllib.request import Request, urlopen
except ImportError:      # pragma: no cover
    # PY2
    from urllib2 import Request, urlopen, HTTPError, URLError

from json import loads

from .dev._exceptions import ISBNLibHTTPError, ISBNLibURLError
from .dev.webservice import query


COVERZOOM = 2
NOIMGSIZE = (15666, 16641, 2690)
UA = "isbnlib (gzip)"

LOGGER = logging.getLogger(__name__)


def bookid(isbn):
    """Return the Google's id associated with each ISBN."""
    from .registry import metadata_cache  # <-- dynamic
    # check the cache fist
    cache = metadata_cache
    if cache is not None:        # pragma: no cover
        key = 'gid' + isbn
        try:                     # pragma: no cover
            if cache[key]:
                return cache[key]
            else:
                raise  # <-- IMPORTANT: usually the caches don't return error!
        except:                  # pragma: no cover
            pass
    url = "https://www.googleapis.com/books/v1/volumes?q="\
          "isbn+{isbn}&fields=items/id&maxResults=1".format(isbn=isbn)
    content = query(url, user_agent=UA)
    try:
        content = loads(content)
        gid = content['items'][0]['id']
        if gid and cache is not None:
            cache[key] = gid
        return gid
    except:                      # pragma: no cover
        return


def _download(url, tofile=None):
    """Download image."""
    headers = {'User-Agent': UA, 'Pragma': 'no-cache'}
    request = Request(url, headers=headers)
    try:
        response = urlopen(request)
        LOGGER.debug('Request headers:\n%s', request.header_items())
    except HTTPError as e:  # pragma: no cover
        LOGGER.critical('ISBNLibHTTPError for %s with code %s [%s]',
                        url, e.code, e.msg)
        if e.code == 403:
            # Google uses this code when the image is not
            # available for any size (should use 404),
            # but also when you are blacklisted by the service
            LOGGER.debug('Cover not available or you are making many requests')
            return True     # <-- no more attempts to download
        if e.code in (401, 429):
            raise ISBNLibHTTPError('%s Are you are making many requests?'
                                   % e.code)
        if e.code in (502, 504):
            raise ISBNLibHTTPError('%s Service temporarily unavailable!'
                                   % e.code)
        raise ISBNLibHTTPError('(%s) %s' % (e.code, e.msg))
    except URLError as e:   # pragma: no cover
        LOGGER.critical('ISBNLibURLError for %s with reason %s',
                        url, e.reason)
        raise ISBNLibURLError(e.reason)
    content = response.read()
    noimageavailable = len(content) in NOIMGSIZE
    if noimageavailable:    # pragma: no cover
        return False
    if tofile:
        try:                # pragma: no cover
            # PY2
            content_type = response.info().getheader('Content-Type')
        except:             # pragma: no cover
            # PY3
            content_type = response.getheader('Content-Type')
        _, ext = content_type.split('/')
        tofile = tofile.split('.')[0] + '.' + ext.split('-')[-1]
        with open(tofile, 'wb') as f:
            f.write(content)
    else:                   # pragma: no cover
        print(content)
    return tofile


def _google_cover(gid, isbn, zoom=COVERZOOM, mode='prt'):
    """Download the cover from Google."""
    tpl = "http://books.google.com/books/content?id={gid}&printsec=frontcover"\
          "&img=1&zoom={zoom}&edge=curl&source=gbs_api"
    url = tpl.format(gid=gid, zoom=zoom)
    if mode == 'url':              # pragma: no cover
        return (url, None)
    coverfile = _download(url, tofile=isbn)
    while not coverfile:           # pragma: no cover
        # try a smaller resolution
        zoom -= 1
        if zoom > 0:
            url = tpl.format(gid=gid, zoom=zoom)
        else:
            return (None, None)
        coverfile = _download(url, tofile=isbn)
    return (url, coverfile) if coverfile and coverfile is not True\
        else (url, None)


def cover(isbn, size=COVERZOOM, mode='prt'):
    """Main entry point for cover."""
    gid = bookid(isbn)
    return _google_cover(gid, isbn, zoom=size, mode=mode) if gid\
        else (None, None)



# -*- coding: utf-8 -*-

"""Registry for metadata services, formatters and cache."""


from . import _goob as goob
from . import _isbndb as isbndb
from . import _merge as merge
from . import _openl as openl
from . import _wcat as wcat
from ._imcache import IMCache
from .dev._fmt import fmtbib

# SERVICES

services = {'default': merge.query,
            'wcat': wcat.query,
            'goob': goob.query,
            'merge': merge.query,
            'openl': openl.query,
            'isbndb': isbndb.query
            }


def setdefaultservice(name):          # pragma: no cover
    """Set the default service."""
    global services
    services['default'] = services[name]


def add_service(name, query):         # pragma: no cover
    """Add a new service to services."""
    global services
    services[name] = query


# FORMATTERS

bibformatters = {'default': lambda x: fmtbib('labels', x),
                 'labels': lambda x: fmtbib('labels', x),
                 'bibtex': lambda x: fmtbib('bibtex', x),
                 'endnote': lambda x: fmtbib('endnote', x),
                 'refworks': lambda x: fmtbib('refworks', x),
                 'msword': lambda x: fmtbib('msword', x),
                 'json': lambda x: fmtbib('json', x),
                 'opf': lambda x: fmtbib('opf', x)}   # pragma: no cover


def setdefaultbibformatter(name):              # pragma: no cover
    """Set the default formatter."""
    global bibformatters
    bibformatters['default'] = bibformatters[name]


def add_bibformatter(name, formatter):         # pragma: no cover
    """Add a new formatter to formatters."""
    global bibformatters
    bibformatters[name] = formatter


# CACHE
# if you want a persistant cache you could use
# .dev.helpers ShelveCache(pathtofile)

metadata_cache = IMCache()            # should be an instance


def set_cache(cache):                 # pragma: no cover
    """Set cache for metadata."""
    global metadata_cache
    metadata_cache = cache

covers_cache = None                   # should be an instance


def set_covers_cache(cache):          # pragma: no cover
    """Set cache for covers."""
    global covers_cache
    covers_cache = cache

custom_cache = None                   # should be an instance


def set_custom_cache(cache):          # pragma: no cover
    """Set a 'spare' cache."""
    global custom_cache
    custom_cache = cache



# -*- coding: utf-8 -*-
"""Process tasks in several modes."""

import logging

from .. import config

LOGGER = logging.getLogger(__name__)


def serial(named_tasks, arg):
    """Serial calls."""
    results = {}
    for name, task in named_tasks:
        try:
            results[name] = task(arg)
        except:    # pragma: no cover
            LOGGER.debug("No result in 'serial' for %s[%s](%s)",
                         task, name, arg)
            results[name] = None
    return results


def parallel(named_tasks, arg):
    """Threaded calls."""
    from threading import Thread
    results = {}

    def _worker(name, task, arg):
        try:
            results[name] = task(arg)
        except:    # pragma: no cover
            LOGGER.debug("No result in 'parallel' for %s[%s](%s)",
                         task, name, arg)
            results[name] = None

    for name, task in named_tasks:
        t = Thread(target=_worker, args=(name, task, arg))
        t.start()
        t.join(config.THREADS_TIMEOUT)
    return results


def multi(named_tasks, arg):
    """Multiprocessing: using several cores (if available)."""
    from multiprocessing import Process, Queue
    results = {}
    q = Queue()

    def _worker(name, task, arg, q):
        try:      # pragma: no cover
            q.put((name, task(arg)))
        except:   # pragma: no cover
            LOGGER.debug("No result in 'multi' for %s[%s](%s)",
                         task, name, arg)
            q.put((name, None))

    for name, task in named_tasks:
        p = Process(target=_worker, args=(name, task, arg, q))
        p.start()
        p.join(config.THREADS_TIMEOUT)
    q.put('STOP')

    while True:
        el = q.get()
        if el == 'STOP':
            break
        results[el[0]] = el[1]
    return results



# -*- coding: utf-8 -*-
"""Query providers for metadata."""

from ._core import EAN13
from ._exceptions import NotRecognizedServiceError, NotValidISBNError
from .registry import services


def query(isbn, service='default', cache='default'):
    """Query worldcat.org, Google Books (JSON API), ... for metadata."""
    from .registry import metadata_cache                  # <-- dinamic now!
    ean = EAN13(isbn)                    # <-- XXX maybe this is too strict?
    if not ean:
        raise NotValidISBNError(isbn)
    isbn = ean
    if cache == 'default':
        cache = metadata_cache
    if service != 'default' and service not in services:  # pragma: no cover
        raise NotRecognizedServiceError(service)
    if cache is None:                                     # pragma: no cover
        return services[service](isbn)
    key = isbn + service
    try:
        if cache[key]:
            return cache[key]
        else:                                             # pragma: no cover
            raise    # <-- IMPORTANT: usually the caches don't return error!
    except:
        meta = services[service](isbn)
        if meta:
            cache[key] = meta
        return meta if meta else None



# -*- coding: utf-8 -*-

import sys


def quiet_errors(exc_type, exc_value, traceback):
    """An error format suitable for end user scripts.

       Usage: enter the following lines in your script
              from isbnlib import quiet_errors
              sys.excepthook = quiet_errors
    """
    sys.stderr.write('Error: %s\n' % exc_value)  # pragma: no cover


class ISBNLibException(Exception):

    """Base class for isbnlib exceptions.

    This exception should not be raised directly,
    only subclasses of this exception should be used!
    """

    def __str__(self):
        return getattr(self, 'message', '')      # pragma: no cover


class NotRecognizedServiceError(ISBNLibException):

    """Exception raised when the service is not in config.py."""

    def __init__(self, service):
        self.message = "(%s) is not a recognized service" % service


class NotValidISBNError(ISBNLibException):

    """Exception raised when the ISBN is not valid."""

    def __init__(self, isbnlike):
        self.message = "(%s) is not a valid ISBN" % isbnlike


class PluginNotLoadedError(ISBNLibException):  # pragma: no cover

    """Exception raised when the plugin's loader doesn't load the plugin.

        TODO: Delete this in version 4
    """

    def __init__(self, path):
        self.message = "plugin (%s) wasn't loaded" % path
